--[[============================================================================
main.lua
============================================================================]]--

--TimeHelper

--for those that want to place events in minute:second -time


--0.2 Initial release






-- Placeholders for the dialogs
local time_dialog = nil
local jump_dialog = nil

-- Placeholders to expose the ViewBuilders outside the show_dialog() functions
local vb_time_dialog = nil
local vb_jump_dialog = nil

-- Reload the script whenever this file is saved. 
-- Additionally, execute the attached function.
_AUTO_RELOAD_DEBUG = function()
    boot() 
end

-- Read from the manifest.xml file.
class "RenoiseScriptingTool" (renoise.Document.DocumentNode)
  function RenoiseScriptingTool:__init()    
    renoise.Document.DocumentNode.__init(self) 
    self:add_property("Name", "Untitled Tool")
    self:add_property("Id", "Unknown Id")
  end

local manifest = RenoiseScriptingTool()
local ok,err = manifest:load_from("manifest.xml")
local tool_name = manifest:property("Name").value
local tool_id = manifest:property("Id").value


local ABS_LINE_NOW --for idle compare
local LINE_NOW
local SEQ_NOW
local UPDATE_FREQ = .1 --seconds
local TIME_NOW = os.clock()

local JUMP_DIALOG_FOCUSED = nil --a global value for storing currently focused textfield in jump dialog
local JUMP_DIALOG_FOCUSED_PREV = nil --a global value for storing previously focused textfield in jump dialog

local PLACE_DELAY_AFTER_JUMP = false --this up when go_to_time() is run -> a delay is placed in song time

--GUI defaults
    local ctrl_margin = renoise.ViewBuilder.DEFAULT_CONTROL_MARGIN
    local ctrl_spacing = renoise.ViewBuilder.DEFAULT_CONTROL_SPACING
    local ctrl_height = renoise.ViewBuilder.DEFAULT_CONTROL_HEIGHT
    local mctrl_height = renoise.ViewBuilder.DEFAULT_MINI_CONTROL_HEIGHT
    local dlg_margin = renoise.ViewBuilder.DEFAULT_DIALOG_MARGIN
    local dlg_spacing = renoise.ViewBuilder.DEFAULT_DIALOG_SPACING
    local dlg_but_height = renoise.ViewBuilder.DEFAULT_DIALOG_BUTTON_HEIGHT

--Booting and exiting
function boot()
    --[[
    Runs when new document is availlable
    --]]
    if not renoise.tool().app_idle_observable:has_notifier(idle_updater)
        then renoise.tool().app_idle_observable:add_notifier(idle_updater)
    end
    return true
end
function exit()
    --[[
    Runs when document is released
    --]]
    if renoise.tool().app_idle_observable:has_notifier(idle_updater)
        then renoise.tool().app_idle_observable:remove_notifier(idle_updater)
    end
    return true
end

if not renoise.tool().app_new_document_observable:has_notifier(boot)
    then renoise.tool().app_new_document_observable:add_notifier(boot)
end
if not renoise.tool().app_release_document_observable:has_notifier(exit)
    then renoise.tool().app_release_document_observable:add_notifier(exit)
end


--Tool GUI values Document
local guivals = renoise.Document.create {
   ['minutes'] = "0", 
   ['seconds'] = "0", 
   ['hundreths'] = "0", 
}


--------------------------------------------------------------------------------
-- Main functions
--------------------------------------------------------------------------------

-- This example function is called from the GUI below.
-- It will return a random string. The GUI function displays 
-- that string in a dialog.
local function get_absolute_line(at_seq, at_line)
    --[[
    Returns the absolute line position. Absolute means 'cumulative' or such.
    --]]
    local lines_total = 0
    for seq_index = 1, at_seq do
        local lines_in_pattern = renoise.song():pattern(renoise.song().sequencer:pattern(seq_index)).number_of_lines
        for line_index = 1, (seq_index == at_seq and at_line) or lines_in_pattern do
            lines_total = lines_total + 1
        end
    end
    lines_total = lines_total - 1 --For getting the time at the START of the line
    --Store line in global for comparison
    ABS_LINE_NOW = lines_total
    --Finished
    return lines_total
end

local function get_timestring(minutes)
    --[[
    Returns a formatted string of minutes:seconds.hundreths
    --]]
    local minutes_full = math.floor(minutes)
    local minutes_frac = minutes - minutes_full
    local seconds = 60 * minutes_frac
    local seconds_full = math.floor(seconds)
    local seconds_frac = seconds - seconds_full
    local hundreths = seconds_frac * 100

    local timestring

    timestring = string.format("%d:%02d.%03d", minutes_full, seconds_full, hundreths)

    return timestring
end
local function get_time(lines_total)
    --[[
    Returns time in minutes with an argument of total line traversal
    TODO:would be super nice if this would account for bpm/lpb changes in the song
    --]]
    --Get beats 
    local bpm = renoise.song().transport.bpm
    local lpb = renoise.song().transport.lpb
    local beats_total = lines_total / lpb
    --Get minutes
    local minutes = beats_total / bpm
    return minutes
end
local function show_time(vb_time_dialog, lines_total)
    --[[
    Shows time in statusbar
    --]]
    local minutes = get_time(lines_total)
    local timestring = "LINE START   "..get_timestring(minutes)
     
    -- Show time
    if vb_time_dialog then
        vb_time_dialog.views['timer_text'].text = timestring
    end
    return true
end

local function go_to_time(minutes_total, place_delay)
    --[[
    Moves the cursor to minutes_total line
    minutes_total = float
    place_delay = boolean
    --]]
    --Get beats 
    local bpm = renoise.song().transport.bpm
    local lpb = renoise.song().transport.lpb
    local beats_total = minutes_total * bpm
    local lines_total = beats_total * lpb
    --Get the place
    local function line_forward(at_seq, at_line)
        --[[
        Moves the target position a single line forward
        --]]
        local this_pattern = renoise.song():pattern(renoise.song().sequencer:pattern(at_seq))
        local lines = this_pattern.number_of_lines
        if at_line > lines then
            error("Strange days have found us")
        end
        if lines == at_line then
            at_seq = at_seq + 1
            at_line = 1
        else
            at_line = at_line + 1
        end

        if at_seq > #renoise.song().sequencer.pattern_sequence then
            --hit the end
            --at_seq = sequence number after line advance
            --at_line = line number after line advance
            --true = still lines left to advance?
            return false
        else
            return true, at_seq, at_line
        end
    end

    local at_seq = 1
    local at_line = 1
    local final_seq = 1
    local final_line = 1
    local success, at_seq, at_line = line_forward(1,0)
    local full_lines = 0
    while success and math.floor(lines_total) > 0 do
        lines_total = lines_total - 1
        full_lines = full_lines + 1
        success, at_seq, at_line = line_forward(at_seq, at_line)
        if success then
            final_seq = at_seq
            final_line = at_line
        end
    end

    --Place cursor
    renoise.song().selected_sequence_index = final_seq
    renoise.song().selected_line_index = final_line

    --calculate offset
    local target_timestring = get_timestring(minutes_total)
    local offset = get_time(lines_total) --lines_total has been reduced to contain the offset
    local offset_timestring = get_timestring(offset)

    local message --the statusbar message to show after operation
    
    if PLACE_DELAY_AFTER_JUMP then
        --Check appropriate track type for delay placing
        local this_track = renoise.song().selected_track
        if this_track.visible_note_columns == 0 then
            message = "Could not place delay, no note columns in track!"
        else
            if lines_total > 1 then
                message = "Could not place delay, target out of reach"
            else
                --Calculate delay value
                local target_delay = lines_total * 256
                local offset_target_delay = target_delay - math.floor(target_delay)
                local offset_target_delay_in_minutes = get_time(1) / 256 * offset_target_delay
                local offset_target_delay_timestring = get_timestring(offset_target_delay_in_minutes)
                local this_pattern_track_line = renoise.song().selected_line
                local this_notecol = renoise.song().selected_note_column or this_pattern_track_line:note_column(this_track.visible_note_columns)
                --Set delay column visible
                this_track.delay_column_visible = true
                --Place delay value
                this_notecol.delay_value = target_delay
                message = "Offset after delay was "..offset_target_delay.." 256ths of a line = "..offset_target_delay_timestring.." in real time. (1 line = "..get_timestring(get_time(1))..", one 256th of a line = "..get_timestring(get_time(1/256))..")"
                --print("Placed a delay of:", string.format("%02X", target_delay))
            end
        end
    else
        message = "Jumped to "..target_timestring..", current line offset from target: "..offset_timestring
    end
    renoise.app():show_status(message)
    return true
end

--------------------------------------------------------------------------------
-- GUI
--------------------------------------------------------------------------------

local function show_jump_dialog()
    --[[
    A GUI for getting a min:sec:hundreths time
    --]]
    -- This block makes sure a non-modal dialog is shown once.
    -- If the jump_dialog is already opened, it will be focused.
    if jump_dialog and jump_dialog.visible then
        jump_dialog:show()
        return
    end

    vb_jump_dialog = renoise.ViewBuilder()

    local function keyhandler(jump_dialog, key)
        --[[
        Keyhandler for GUI get_time
        --]]
        --print("Keyhandler received key.name:"..key.name.." key.modifiers:"..key.modifiers)
        --Map actions to keys
        local actions = {
            ['CLOSE DIALOG'] = {"esc", ""},
        }
        --Test if reserved
        local this_action = ""
        local is_reserved_key = false
        for action, res_key in pairs(actions) do
            if key.name == res_key[1] and key.modifiers == res_key[2] then
                is_reserved_key = true
                this_action = action
                break
            end
        end
        if not is_reserved_key then
            --Send key back to renoise (Global context)
            return key
        end
        --Do actions
        if this_action == "CLOSE DIALOG" then
            --set 
            --JUMP_DIALOG_ON = false
            JUMP_DIALOG_FOCUSED = nil
            JUMP_DIALOG_FOCUSED_PREV = nil
            jump_dialog:close()
            jump_dialog = nil
            vb_jump_dialog = nil
        elseif this_action == "ACTION1" then
            --Not there yet
        else
            error("Unidentified action for tool GUI: "..(action or "nil/false"))
        end
    end
    
    local dlg_wd = 120
    local dlg_wd_effective = dlg_wd - 2 * dlg_margin
    local part_wd = dlg_wd_effective / 3
    local part_wd_effective = dlg_wd_effective - 2 * dlg_spacing
    local content = vb_jump_dialog:row{
        id = "get_time_row",
        margin = dlg_margin,
        spacing = dlg_spacing,
        height = dlg_but_height + 2 * dlg_margin,
        vb_jump_dialog:column{
            id = "get_time_minutes_column",
            margin = 0,
            spacing = 0,
            width = part_wd,
            vb_jump_dialog:textfield{
                id = "minutes_textfield",    
                align = "left",
                width = part_wd_effective,
                bind = guivals.minutes,
            },
        },
        vb_jump_dialog:column{
            id = "get_time_seconds_column",
            margin = 0,
            spacing = 0,
            width = part_wd,
            vb_jump_dialog:textfield{
                id = "seconds_textfield",    
                align = "left",
                width = part_wd_effective,
                bind = guivals.seconds,
            },
        },
        vb_jump_dialog:column{
            id = "get_time_hundreths_column",
            margin = 0,
            spacing = 0,
            width = part_wd,
            vb_jump_dialog:textfield{
                id = "hundreths_textfield",    
                align = "left",
                width = part_wd_effective,
                bind = guivals.hundreths,
            },
        }
    }
    
    local title
    if PLACE_DELAY_AFTER_JUMP then
        title = "Place delay at time"
    else
        title = "Jump to time"
    end
    jump_dialog = renoise.app():show_custom_dialog(title, content, keyhandler)  
    --Set global
    --JUMP_DIALOG_ON = true

    return true
end
local function show_time_dialog()

    -- This block makes sure a non-modal dialog is shown once.
    -- If the time_dialog is already opened, it will be focused.
    if time_dialog and time_dialog.visible then
        time_dialog:show()
        return
    end

    local function keyhandler(time_dialog, key)
        --[[
        Keyhandler for GUI timer
        --]]
        --print("Keyhandler received key.name:"..key.name.." key.modifiers:"..key.modifiers)
        --Map actions to keys
        local actions = {
            ['CLOSE DIALOG'] = {"esc", ""},
            ['GO TO TIME'] = {"g", "shift"},
            ['GO TO TIME AND PLACE DELAY'] = {"d", "shift"},
        }
        --Test if reserved
        local this_action = ""
        local is_reserved_key = false
        for action, res_key in pairs(actions) do
            if key.name == res_key[1] and key.modifiers == res_key[2] then
                is_reserved_key = true
                this_action = action
                break
            end
        end
        if not is_reserved_key then
            --Send key back to renoise (Global context)
            return key
        end
        --Do actions
        if this_action == "CLOSE DIALOG" then
            time_dialog:close()
            time_dialog = nil
            vb_time_dialog = nil
        elseif this_action == "GO TO TIME" then
            PLACE_DELAY_AFTER_JUMP = false
            show_jump_dialog(false)
        elseif this_action == "GO TO TIME AND PLACE DELAY" then
            PLACE_DELAY_AFTER_JUMP = true
            show_jump_dialog(true)
        else
            error("Unidentified action for tool GUI: "..(action or "nil/false"))
        end

    end

    -- The ViewBuilder is the basis
    vb_time_dialog = renoise.ViewBuilder()

    -- The content of the time_dialog, built with the ViewBuilder.
    local content = vb_time_dialog:column{
        id = "timer_column",
        margin = dlg_margin,
        spacing = dlg_spacing,
        width = 200,
        vb_time_dialog:text{
            id = "timer_text",    
            font = "normal",
            align = "left",
            text = "Initializing...",
        }
    }

    -- A custom dialog is non-modal and displays a user designed
    -- layout built with the ViewBuilder.   
    time_dialog = renoise.app():show_custom_dialog("Line start time", content, keyhandler)  
    --Set global
    --TIME_DIALOG_ON = true
end

function idle_updater()
    --[[
    This runs in the background, checks stuff
    --]]
    if os.clock() - TIME_NOW > UPDATE_FREQ then
        TIME_NOW = os.clock()
        --Time update on time dialog
        if time_dialog and time_dialog.visible then
            --Check if line number is different than before
            local at_line = renoise.song().selected_line_index
            local at_seq = renoise.song().selected_sequence_index
            local lines_total = get_absolute_line(at_seq, at_line)
            if LINE_NOW == at_line and SEQ_NOW == at_seq then
                --No change
            else
                --If it is, show time
                local lines_total = get_absolute_line(at_seq, at_line)
                show_time(vb_time_dialog, lines_total)
                --Update globals
                SEQ_NOW = at_seq
                LINE_NOW = at_line
                --Finished
            end
        end
        --Textfield focus checking on jump dialog
        if jump_dialog and jump_dialog.visible then
            local function focuscheck()
                --[[
                populates the JUMP_DIALOG_FOCUSED with the vb.view that's focused
                --]]
                if vb_jump_dialog.views['minutes_textfield'].edit_mode == true then
                    JUMP_DIALOG_FOCUSED = "minutes"
                elseif vb_jump_dialog.views['seconds_textfield'].edit_mode == true then
                    JUMP_DIALOG_FOCUSED = "seconds"
                elseif vb_jump_dialog.views['hundreths_textfield'].edit_mode == true then
                    JUMP_DIALOG_FOCUSED = "hundreths"
                else
                    JUMP_DIALOG_FOCUSED = nil
                end
                return true
            end
            --Check if some field is focused
            if not JUMP_DIALOG_FOCUSED then
                --Nothing's focused
                if not JUMP_DIALOG_FOCUSED_PREV then
                    --It's not focused yet. Start by focusing minutes.
                    vb_jump_dialog.views['minutes_textfield'].edit_mode = true
                    JUMP_DIALOG_FOCUSED = "minutes"
                    JUMP_DIALOG_FOCUSED_PREV = "minutes" --illogical as it sounds...
                elseif JUMP_DIALOG_FOCUSED_PREV == "minutes" then
                    --Minutes was just focused
                    --Check minutes value
                    if not tonumber(guivals.minutes.value) then
                        guivals.minutes.value = "0"
                    end
                    vb_jump_dialog.views['seconds_textfield'].edit_mode = true
                    JUMP_DIALOG_FOCUSED = "seconds"
                    JUMP_DIALOG_FOCUSED_PREV = "seconds"
                elseif JUMP_DIALOG_FOCUSED_PREV == "seconds" then
                    --Seconds was just focused
                    --Check seconds value
                    if not tonumber(guivals.seconds.value) then
                        guivals.seconds.value = "0"
                    elseif tonumber(guivals.seconds.value) > 59 then
                        guivals.seconds.value = "59"
                    end
                    vb_jump_dialog.views['hundreths_textfield'].edit_mode = true
                    JUMP_DIALOG_FOCUSED = "hundreths"
                    JUMP_DIALOG_FOCUSED_PREV = "hundreths"
                elseif JUMP_DIALOG_FOCUSED_PREV == "hundreths" then
                    --Hundreths was just focused
                    --Check hundreths value
                    if not tonumber(guivals.hundreths.value) then
                        guivals.hundreths.value = "0"
                    elseif tonumber(guivals.hundreths.value) > 99 then
                        guivals.hundreths.value = "99"
                    end
                    JUMP_DIALOG_FOCUSED = nil
                    JUMP_DIALOG_FOCUSED_PREV = nil
                    --All is done
                    -- Deduce minutes_total
                    local minutes_total
                    minutes_total = tonumber(guivals.minutes.value) or 0
                    minutes_total = minutes_total + (((tonumber(guivals.seconds.value) or 0) + (tonumber(guivals.hundreths.value)*0.01 or 0) )/ 60)
                    go_to_time(minutes_total, PLACE_DELAY_AFTER_JUMP)
                    jump_dialog:close()
                    jump_dialog = nil
                    vb_jump_dialog = nil
                else
                    error("Weirdness ahoy_1")
                end
            else
                --Something's focused
                if JUMP_DIALOG_FOCUSED == "minutes" then
                    --Minutes field is focused
                    focuscheck()
                elseif JUMP_DIALOG_FOCUSED_PREV == "seconds" then
                    --Seconds field is focused
                    focuscheck()
                elseif JUMP_DIALOG_FOCUSED_PREV == "hundreths" then
                    --hundreths field is focused
                    focuscheck()
                else
                    error("Weirdness ahoy_2")
                end
            end

        end
    end
end

--------------------------------------------------------------------------------
-- Menu entries
--------------------------------------------------------------------------------

renoise.tool():add_menu_entry {
  name = "Main Menu:Tools:TimeHelper:Jump cursor to time...",
  invoke = function()
     PLACE_DELAY_AFTER_JUMP = false 
     show_jump_dialog()
  end
}
renoise.tool():add_menu_entry {
  name = "Main Menu:Tools:TimeHelper:Place delay marker at time...",
  invoke = function()
     PLACE_DELAY_AFTER_JUMP = true
     show_jump_dialog()
  end
}


--------------------------------------------------------------------------------
-- Key Binding
--------------------------------------------------------------------------------

renoise.tool():add_keybinding {
  name = "Global:Tools:Time Helper...",
  invoke = show_time_dialog
}


--------------------------------------------------------------------------------
-- MIDI Mapping
--------------------------------------------------------------------------------

--[[
renoise.tool():add_midi_mapping {
  name = tool_id..":Show Dialog...",
  invoke = show_time_dialog
}
--]]
